#ifndef DS_H
#define DS_H

#include <stdbool.h>

/* Constants */

#define INITITEMS 10
#define TRUE 1
#define FALSE 0
#define INF 1.0E300
#define HASHSIZE 10000

/* Data structure definitions */

typedef struct vertex {

  int maxedges; /* Indicates how much space has been allocated for incidentedges */

  struct edge **incidentedges;
  int degree;
  int visited;
  int vindex;
  
  double rank; // useful for sorting vertices, e.g. for Dijkstra distance algo.

} vertex;

typedef struct edge {

  struct vertex *vertices[2];
  double weight;
  int traversed;

} edge;

typedef struct graph {

  struct vertex **vertices;
  struct edge **edgelist;
  
  char **labels;
  
  int nvertices;
  int nedges;
  
  int maxvertices;
  int maxedges;

} graph;


typedef struct hashitem {

  void *key;
  void *value;
  struct hashitem *next;

} hashitem;

typedef struct hashtable {

  struct hashitem **table;
  int (*hash)(const void*);
  int (*compar)(const void*, const void*);
  void (*freekey)(void*);
  void (*freeval)(void*);
  int tablesize;
  int size; // number of items
  
  struct hashitem *iter;

} hashtable;

typedef struct heapnode {
  
  struct heapnode *left;
  struct heapnode *right;
  struct heapnode *parent;
  //int rank; // Do I want this? Or do I want to require a compar function?
  int index;
  
  const void *item;
  
} heapnode;

typedef struct heap {
  
  int (*compar)(const void*, const void*);
  int size;
  int maxsize;

  struct heapnode **array;

} heap;

typedef struct llnode {
  
  struct llnode *next;
  struct llnode *last;
  
  void *item;
  
} llnode;

typedef struct linkedlist {

  struct llnode *head;
  struct llnode *tail;
  struct llnode *iter;
  
} linkedlist;

typedef struct rbnode {
  
  void *item;
  bool colour;
  
  struct rbnode *parent;
  struct rbnode *left;
  struct rbnode *right;
  
} rbnode;

typedef struct rbtree {
  
  struct rbnode *root;
  int (*compar)(const void*, const void*);

} rbtree;

/* treenode type is same as llnode type, but strung together differently */
typedef struct treenode {
  
  struct treenode *next;
  struct treenode *out;
  double brlen;
  char *label;
  int index;
  
  struct treebipart *binrep;
  
} treenode; 

typedef struct tree {
  
  int nleaves;
  int ninternal;
  char **leafnames;
  struct treenode **nodearray;
  struct treenode *root;
  
  struct treenode *spareleaves;
  
  char *treestr;
  //int trstrlen;

} tree;

typedef struct treebipart {
  
  int nleaves;
  bool *array;
  
} treebipart;

typedef struct vector {
  int size;
  int maxsize;

  void **items;

} vector;


/* Function prototypes */

/* from graph.c */
graph* testgraph4();
graph* initgraph(graph*, int, int);
void freegraph(graph*);
vertex* addvertexgraph(graph*, char*);
edge* addedgegraph(graph*, vertex*, vertex*, double);
vertex** Dijkstragraph(graph*, vertex*, vertex**, double*, double, bool);
void centralitiesgraph(graph*, double*, double*, double*);
void printgraph(graph*);
graph* buildgraph(graph*, double*, char**, int, bool);
bool isconnected(graph *);

/* from hashtable.c */
hashtable* inithashtable(hashtable*, int(*)(const void*), int(*)(const void*, const void*), void(*)(void*), void(*)(void*), int);
void freehashtable(hashtable*);
void inserthashtable(hashtable*, void*, void*);
void* removehashtable(hashtable*, const void*);
void* gethashtable(hashtable*, const void*);
void resetiterhashtable(hashtable*);
const void* nextkeyhashtable(hashtable*);
void* nextvaluehashtable(hashtable*);


/* from heap.c */
heap* initheap(heap*, int(*)(const void*, const void*));
void freeheap(heap*);
void addheap(heap*, const void*);
const void* removeheap(heap*);
void bubbleupheap(heap*);
void bubbledownheap(heap*);
void swapheap(heap*, heapnode*, heapnode*);

/* from ll.c */
linkedlist* initlinkedlist(linkedlist*);
void freelinkedlist(linkedlist*);
void addlinkedlist(linkedlist*, void*);
void* removelinkedlist(linkedlist*, void*);
void resetlinkedlist(linkedlist*);
void *nextlinkedlist(linkedlist*);
void sortlinkedlist(llnode*, llnode*, int(*)(const void*, const void*));
llnode* partitionlinkedlist(llnode*, llnode*, int(*)(const void*, const void*));
void swaplinkedlist(llnode*, llnode*);
void enqueuelinkedlist(linkedlist*, void*);
void* dequeuelinkedlist(linkedlist*);
void pushlinkedlist(linkedlist*, void*);
void* poplinkedlist(linkedlist*);

/* from rbtree.c */
rbtree* initrbtree(rbtree *t, int (*)(const void*, const void*));
rbnode* initrbnode(rbnode*, void*);
rbnode* initsentinelrbnode(rbnode*);
rbnode* grandparentrbnode(rbnode*);
rbnode* siblingrbnode(rbnode*);
rbnode* unclerbnode(rbnode*);
rbnode* insertitemrbtree(rbtree*, void*);
void fixcoloursrbnode(rbnode*);
void rotateright(rbnode*);
void rotateleft(rbnode*);
void* getitemrbtree(rbtree*, void*);
void* removeitemrbtree(rbtree*, void*);
void removerbnode(rbnode*);
void deleterbnode(rbnode*);
void replacerbnode(rbnode*, rbnode*);
rbnode* findmaxrbnode(rbnode*);
int isleaf(rbnode*);
void freerbtree(rbtree*);
void freerbnodeREC(rbnode*);

/* from tree.c */
treenode* inittreenode(treenode*, int, int);
void resettreenode(treenode*, int, int);
void freetreenode(treenode*);
treebipart* inittreebipart(treebipart*, int, int);
void resettreebipart(treebipart*, int, int);
void freetreebipart(treebipart*);
void freevoidtreebp(void*);
treebipart* copytreebipart(treebipart*, treebipart*);
treebipart* combinetreebipart(treebipart*, treebipart*);
tree* inittree(tree*, char**, int);
tree* freetree(tree*);
tree* string2tree(tree*, char*);
tree* restring2tree(tree*, char*);
//tree* string2unrestree(tree*, char*, double);
//tree* restring2unrestree(tree*, char*, double);
treenode* remtrivialnodetree(tree*, treenode*);
void collapsebranchtree(treenode*);
void collapseshortbranchestree(tree*, double);
void collapseshortbranchestreeREC(tree*, treenode*, double);
bool connectleaftree(tree*, treenode*, char*);
bool setbrlentreenode(treenode*, double);
char* makestringtree(tree*);
void makestringtreeREC(tree*, treenode*);
void printtree(tree*);
hashtable* countbipartstree(tree*, hashtable*);//double*);
void countbipartstreeREC(tree *t, treenode*, hashtable*); //double*); /***/
void printbipartstree(tree*);
static int cmptreenode(const void*, const void*);
int hashtreebipart(const void*);
int cmptreebipart(const void*, const void*);
treenode* findnodetree(tree*, char*, int, int);
void fixleavestree(tree*);

/* from vector.c */
vector* initvector(vector*);
void freevector(vector*);
void* getvector(vector*, int);
void* removevector(vector*, int);
void addvector(vector*, void*);
int sizevector(vector*);

#endif
