/**
 * hashtable.c
 * A hash table implementation that uses chaining to resolve collisions.
 * Note that keys MUST be some sort of integer. Consider casting pointer 
   values to int, if exact comparison by reference is desired
 * (c) 2009 JW Leigh
 */
 
#include <stdlib.h>
#include <stdbool.h>
#include <stdio.h>
#include "ds.h"
 
hashtable* inithashtable(hashtable *ht, int (*hash)(const void*), int (*compar)(const void*, const void*), void (*freekey)(void *), void (*freeval)(void *), int tablesize)
{
  int i;
  ht->tablesize = tablesize;
  ht->table = (hashitem**)malloc(tablesize * sizeof(hashitem*));
  
  for (i = 0; i < tablesize; i++)  ht->table[i] = NULL;
  ht->hash = hash;
  ht->compar = compar;
  ht->freekey = freekey;
  ht->freeval = freeval;
  ht->size = 0;
  
  return ht;
}

void freehashtable(hashtable *ht)
{
  int i;
  hashitem *tmp, *tmpnxt;
  
  for (i = 0; i < ht->tablesize; i++)
  {
    tmp = ht->table[i];
    if (tmp)
    {
      while (tmp->next)
      {
        tmpnxt = tmp->next;
        if (ht->freekey)  ht->freekey(tmp->key);
        free(tmp->key);
        
        if (ht->freeval)  ht->freeval(tmp->value);
        free(tmp->value);
        
        free(tmp);
        tmp = tmp->next;
      }
    
      free(tmp);
      //printf("item left at index %d!\n", i);
      
      //fprintf(stderr, "Empty the hashtable before freeing it, or horrible memory leaks will result!\n");
    }
  }
  
  free(ht->table);
}

void inserthashtable(hashtable *ht, void *key, void *value)
{
  int index = ht->hash(key);
  hashitem *tmp;
  
  if (index >= ht->tablesize)
  {
    fprintf(stderr, "Error! Hash function produced an out-of-range index!\n");
    exit(1);
  }
  
  if (ht->table[index])
  {
    tmp = ht->table[index];
    while (tmp->next)  tmp = tmp->next;
    
    tmp->next = (hashitem*)malloc(sizeof(hashitem));
    tmp = tmp->next;
  }
  
  else
  {
    tmp = (hashitem*)malloc(sizeof(hashitem));
    ht->table[index] = tmp;
  }
  
  tmp->key = key;
  tmp->value = value;
  tmp->next = NULL;
  
  ht->size++;
} 
/******************************************************
 * Check whether key exists before inserting!!!!!!!!! *
 ******************************************************/

/**
 * Return NULL if key does not exist
 */
void* removehashtable(hashtable *ht, const void *key)
{
  hashitem *tmp, *prev = NULL;
  bool keyfound = FALSE;
  void *value;
  int index;
  
  index = ht->hash(key);
  
  tmp = ht->table[index];
  
  if (!tmp)  return NULL;
  
  keyfound = !(ht->compar(&(tmp->key), &key));
  
  while (tmp->next && !keyfound)
  {
    prev = tmp;
    tmp = tmp->next;
    keyfound = !(ht->compar(&(tmp->key), &key));
  }
    
  if (!keyfound)  return NULL;
  
  if (prev)  prev->next = tmp->next;
  else  
  //{
    //printf("Nothing left at index %d!\n", index);
    ht->table[index] = NULL;
  //}
  
  value = tmp->value;
  ht->freekey(tmp->key);
  free(tmp->key);
  free(tmp);
  
  ht->size--;
  return value;
}


/**
 * Return NULL if key does not exist
 */
void* gethashtable(hashtable *ht, const void *key)
{
  hashitem *tmp;
  bool keyfound = FALSE;
  void *value;
  int index;
  
  index = ht->hash(key);
  
  tmp = ht->table[index];
  
  if (!tmp)  return NULL;
  
  keyfound = !(ht->compar(&(tmp->key), &key));
  
  while (tmp->next && !keyfound)//tmp->key != key)  tmp = tmp->next;
  {
    tmp = tmp->next;
    keyfound = !(ht->compar(&(tmp->key), &key));
  }
  
  if (!keyfound)  return NULL;//(tmp->key != key)  return NULL;
      
  return tmp->value;
}

void resetiterhashtable(hashtable *ht)
{
  int i;
  ht->iter = NULL;
  for (i = 0; i < ht->tablesize && !ht->iter; i++)
    ht->iter = ht->table[i];
}

const void* nextkeyhashtable(hashtable *ht)
{
  int index, i; 
  void *key;
  
  if (!ht->iter)  return NULL;
  
  key = ht->iter->key;
  index = ht->hash(key);
  
  ht->iter = ht->iter->next;
  for (i = index + 1; i < ht->tablesize && !ht->iter; i++)
    ht->iter = ht->table[i];
  
  return key;
}

/*long* keyshashtable(hashtable *ht, long* keyarray)
{
  int index, i = 0, j;
  hashitem *iter;
  
  for (j = 0; j < ht->tablesize; j++)
  {
    iter = ht->table[j];
    while (iter)
    //for (iter = ht->table[j]; iter; iter = iter->next)
    {
      //if (iter)  {printf("iter is true. Value: %d\n", iter); printf("  key: %d\n", iter->key);}
      
      keyarray[i] = iter->key;
      i++;
      iter = iter->next;
    }
  }
  
  return keyarray;
}*/

void *nextvaluehashtable(hashtable *ht)
{
  int index, i; 
  void *value;
  
  if (!ht->iter)  return NULL;
  
  value = ht->iter->value;
  index = ht->hash(ht->iter->key);
  
  ht->iter = ht->iter->next;
  for (i = index + 1; i < ht->tablesize && !ht->iter; i++)
    ht->iter = ht->table[i];
  
  return value;
}
