/**
 * heap.c
 * A heap data structure
 * (c) 2009 JW Leigh
 */
#include <stdlib.h>
#include <stdio.h>
#include "ds.h"

heap* initheap(heap *H, int(*compar)(const void*, const void*))
{
  int i;
  heapnode *hn;
  H->array = (heapnode**)malloc(INITITEMS * sizeof(heapnode*));
  
  for (i = 0; i < INITITEMS; i++)  
  {
    H->array[i] = (heapnode*)malloc(sizeof(heapnode));
    hn = H->array[i];
    hn->parent = hn->left = hn->right = NULL;
  }
  
  H->size = 0;
  H->maxsize = INITITEMS;
  H->compar = compar;
  
  return H;
}

void freeheap(heap *H)
{
  int i;
  
  for (i = 0; i < H->maxsize; i++)  free(H->array[i]);
  
  free(H->array);
}

void addheap(heap *H, const void *item)
{
  int i;
  heapnode *parent;
  heapnode *hn;
  
  if (H->size >= H->maxsize)
  {
    H->array = (heapnode**)realloc(H->array, H->maxsize * 2 * sizeof(heapnode*));
    
    for (i = H->maxsize; i < H->maxsize * 2; i++)
    {
      H->array[i] = (heapnode*)malloc(sizeof(heapnode));
      hn = H->array[i];
      hn->parent = hn->left = hn->right = NULL;
    }
      
    H->maxsize *= 2;
  }
  
  hn = H->array[H->size];//(*heapnode)malloc(sizeof(heapnode));
  //hn->rank = rank;
  hn->item = item;
  hn->index = H->size;
  
  if (hn->index > 0)
  {
    parent = H->array[(hn->index - 1)/2];
    hn->parent = parent;
    
    /* If index is even, hn is a right node */
    if (hn->index / 2 * 2 == hn->index)
    {
      if (parent->right)
      {
        fprintf(stderr, "Error! Parent already has a right child!");
        exit(1);
      }
      parent->right = hn;
    }
    
    else
    {
      if (parent->left)
      {
        fprintf(stderr, "Error! Parent already has a left child!");
        exit(1);
      }
      parent->left = hn;
    }
  }
  
  H->size++;
  bubbleupheap(H);
}

const void* removeheap(heap *H)
{
  const void *item;
  heapnode *hn;
  
  if (H->size == 0)  return NULL;
  
  H->size--;
  hn = H->array[H->size];
  
  if (hn->parent)
  {
    if (hn->index / 2 * 2 == hn->index)
      hn->parent->right = NULL;
    else
      hn->parent->left = NULL;
  }
  hn->parent = NULL;
  
  item = H->array[0]->item;  
  H->array[0]->item = hn->item;
  //H->array[0]->rank = hn->rank;
  
  hn->item = NULL;
  
  if (H->size > 0)  bubbledownheap(H);
  
  return item;
}

void bubbleupheap(heap *H)
{
  heapnode *hn = H->array[H->size - 1];
  
  while (hn->parent && H->compar(hn->item, hn->parent->item) < 0)//hn->rank < hn->parent->rank)
  {
    swapheap(H, hn, hn->parent);
    hn = hn->parent;
  }
}

void bubbledownheap(heap *H)
{
  heapnode *hn = H->array[0];
  int leftresult, rightresult;
  
  if (hn->left)  leftresult = H->compar(hn->item, hn->left->item);
  if (hn->right)  rightresult = H->compar(hn->item, hn->right->item);
  
  while ((hn->left && leftresult > 0) ||
         (hn->right && rightresult > 0))
          //hn->rank > hn->left->rank) || 
          //(hn->right && hn->rank > hn->right->rank))
  {
    if (!hn->right || H->compar(hn->left->item, hn->right->item) <= 0)
    //if (!hn->right || hn->left->rank <= hn->right->rank)
    {
      swapheap(H, hn, hn->left);
      hn = hn->left;
    }
    else
    {
      swapheap(H, hn, hn->right);
      hn = hn->right;
    }
    
    if (hn->left)  leftresult = H->compar(hn->item, hn->left->item);
    if (hn->right)  rightresult = H->compar(hn->item, hn->right->item);
  }
}

void swapheap(heap *H, heapnode *hn, heapnode *hm)
{
  //int tmp;
  const void *tmpptr;
  heapnode *parent, *child;
  
  if (hn->index < hm->index)
  {
    parent = hn;
    child = hm;
  }
  
  else
  {
    parent = hm;
    child = hn;
  }
  
  if (child->parent != parent)
  {
    fprintf(stderr, "Can only swap parent and child in heap!\n");
    exit(1);
  }
  
  /* Don't swap nodes, just contents */
  //tmp = parent->rank;
  //parent->rank = child->rank;
  //child->rank = tmp;
  
  tmpptr = parent->item;
  parent->item = child->item;
  child->item = tmpptr;
}


