/**
 * ll.c
 * A double-linked list data structure
 * (c) 2009 JW Leigh
 */
#include <stdlib.h>
#include "ds.h"

linkedlist* initlinkedlist(linkedlist *ll)
{
  ll->head = ll->tail = ll->iter = NULL;
  
  return ll;
}

void freelinkedlist(linkedlist *ll)
{
  
  ll->iter = ll->head;
  
  while(ll->iter)
  {
    ll->iter = ll->iter->next;
    free(ll->head);
    ll->head = ll->iter;
  }
}

void addlinkedlist(linkedlist *ll, void *item)
{
  llnode* newnode = (llnode*)malloc(sizeof(llnode));
  
  newnode->item = item;
  newnode->next = newnode->last = NULL;
  
  if (ll->tail)
  {
    ll->tail->next = newnode;
    newnode->last = ll->tail;
    ll->tail = newnode;
  }
  
  else  ll->head = ll->tail = ll->iter = newnode;
}

void* removelinkedlist(linkedlist *ll, void *item)
{
  llnode *next, *last;
  ll->iter = ll->head;
  
  while(ll->iter && ll->iter->item != item)  ll->iter = ll->iter->next;

  if (!ll->iter)  return NULL;
  
  last = ll->iter->last;
  next = ll->iter->next;
  
  if (last)  last->next = next;
  else  ll->head = next;
  
  if (next)  next->last = last;
  else  ll->tail = last;
  
  free(ll->iter);
  
  ll->iter = ll->head;
  
  return item;
}

void resetlinkedlist(linkedlist *ll)
{
  ll->iter = ll->head;
}

void* nextlinkedlist(linkedlist *ll)
{
  llnode *node = ll->iter;
  
  if (ll->iter)
  {
    ll->iter = ll->iter->next;
    return node->item;
  }
  return NULL;
}

/**
 * Quicksort function.
 * Not too smart: doesn't deal with "worst case". Chooses the last node in list
 * and subsequent sublists as pivot points.
 */
void sortlinkedlist(llnode *start, llnode *end, int(*compar)(const void *, const void *))
{  
  llnode *pivot;
  
  if (start && end && start != end)  
  {
    pivot = partitionlinkedlist(start, end, compar);
    
    if (pivot != start)  sortlinkedlist(start, pivot->last, compar);
    if (pivot != end)  sortlinkedlist(pivot->next, end, compar);
  }
}

llnode* partitionlinkedlist(llnode *start, llnode *end, int(*compar)(const void *, const void *))
{
  void *x = end->item;
  llnode *n, *m;
  
  n = start->last;
  m = start;
  
  while (m != end)
  {
    if (compar(m->item, x) <= 0)
    {
      if (n)  n = n->next;
      else  n = start;
      swaplinkedlist(n, m);
    }
    m = m->next;
  }
  
  if (n)
  {
    swaplinkedlist(n->next, end);
    return n->next;
  }
  
  /* end's place is at the start of this list or sublist */
  swaplinkedlist(start, end);
  return start;
  
}

void swaplinkedlist(llnode *n, llnode *m)
{
  void *item = n->item;
  n->item = m->item;
  m->item = item;
}


void enqueuelinkedlist(linkedlist *ll, void *item)
{
  addlinkedlist(ll, item);
}

void* dequeuelinkedlist(linkedlist *ll)
{
  void *item;
  llnode *node;
  
  if (!ll->head)  return NULL;
  
  item = ll->head->item;
  node = ll->head;
  ll->head = ll->head->next;
  
  if (!ll->head)  ll->tail = ll->iter = NULL;
  
  free(node);
  
  return item;
}

void pushlinkedlist(linkedlist *ll, void *item)
{
  addlinkedlist(ll, item);
}

void* poplinkedlist(linkedlist *ll)
{
  void *item;
  llnode *node;
  
  if (!ll->tail)  return NULL;
  
  item = ll->tail->item;
  node = ll->tail;
  ll->tail = ll->tail->last;
  
  if (!ll->tail)  ll->head = ll->iter = NULL;
  
  free(node);
  
  return item;
}
