/**
 * rbtree.c
 * Red-Black tree implementation.
 * (c) 2009 JW Leigh
 */

#include <stdlib.h>
#include <stdio.h>
#include "ds.h"

#define RED 1
#define BLACK 0
rbtree* initrbtree(rbtree *t, int (*compar)(const void*, const void*))
{
  t->root = NULL;
  t->compar = compar;
  
  return t;
}

rbnode* initrbnode(rbnode *rbn, void *item)
{
  rbn->left = initsentinelrbnode(rbn);
  rbn->right = initsentinelrbnode(rbn);
  
  rbn->colour = RED;
  rbn->item = item;
}

rbnode* initsentinelrbnode(rbnode *parent)
{
  rbnode *sentinel = (rbnode*)malloc(sizeof(rbnode));
  sentinel->colour = BLACK;
  sentinel->item = NULL;
  sentinel->parent = parent;
  sentinel->left = sentinel->right = NULL;
  
  return sentinel;
}

rbnode* grandparentrbnode(rbnode *rbn)
{  
  if (!rbn->parent)  return NULL;
  return rbn->parent->parent;
  //(rbn->parent) ? return rbn->parent->parent : return  NULL; 
}

rbnode* siblingrbnode(rbnode *rbn)
{
  if (!rbn->parent)  return NULL;
  
  if (rbn->parent->right == rbn)  return rbn->parent->left;
  return rbn->parent->right;
  //(rbn->parent->right == rbn) ? 
  //    return rbn->parent->left :
  //    return rbn->parent->right;
}

rbnode* unclerbnode(rbnode *rbn)
{
  rbnode* g = grandparentrbnode(rbn);
  
  if (!g) return NULL;
    
  if (rbn->parent == g->left)  return g->right;
  
  if (rbn->parent == g->right)  return g->left;
  
  fprintf(stderr, "Error! Pointers have gotten mixed up, grandparent doesn't have parent as its child.\n");
  
  return NULL;
}

// need some sort of function pointer to compare items 

rbnode* insertitemrbtree(rbtree *t, void *item)
{
  rbnode *iter, *u, *g;
  int result;
   
  if (!(t->root))
  {
    t->root = (rbnode*)malloc(sizeof(rbnode));
    initrbnode(t->root, item);
    t->root->parent = NULL;
    
    t->root->colour = BLACK;
    return t->root;
  }
  
  iter = t->root;
  while (iter->item)
  {
    result = t->compar(item, iter->item);
    if (result < 0)  iter = iter->left;
    else if (result > 0)  iter = iter->right;
    
    /* found key, replace item and return */
    else  
    {
      iter->item = item;
      return iter;
    }
  }
  
  initrbnode(iter, item);
  /*if (iter->parent->colour == BLACK)  return iter;
  
  u = uncle(iter);
  if (u && u->colour == RED)
  {
    
  }*/
  fixcoloursrbnode(iter);
  
  while (t->root->parent)  t->root = t->root->parent;
  t->root->colour = BLACK;
}

void fixcoloursrbnode(rbnode *rbn)
{
  rbnode *u, *g;
  /* rbn is the root of the tree */
  if (rbn->parent == NULL)  
  {
    rbn->colour = BLACK;
    return;
  }
  
  if (rbn->parent->colour == BLACK)  return;

  u = unclerbnode(rbn);
  
  if (u && u->colour == RED)
  {
    rbn->parent->colour = BLACK;
    u->colour = BLACK;
    g = grandparentrbnode(rbn);
    g->colour = RED;
    fixcoloursrbnode(g);
  }
  
  else
  {
    g = grandparentrbnode(rbn);
    
    if (rbn == rbn->parent->right && rbn->parent == g->left)
    {
      rotateleft(rbn->parent); // make this function return node rotated?
      rbn = rbn->left;
    }
      
    else if (rbn == rbn->parent->left && rbn->parent == g->right)
    {
      rotateright(rbn->parent);
      rbn = rbn->right;
    }
    // is it possible for rbn to not have a grandparent?
    rbn->parent->colour = BLACK;
    //g = grandparent(rbn);
    g->colour = RED;
    
    if (rbn == rbn->parent->left)
      rotateright(g);
    
    if (rbn == rbn->parent->right)
      rotateleft(g);
  }
}

void rotateright(rbnode *rbn)
{
  rbnode *parent, *left;
  
  parent = rbn->parent;
  left = rbn->left;
  
  /* note that left will never be NULL */
  rbn->parent = left;
  rbn->left = left->right;
  rbn->left->parent = rbn;
  
  left->right = rbn;
  left->parent = parent;
  
  if (parent)
  {
    if (rbn == parent->left)  parent->left = left;
    else  parent->right = left;
  }
}

void rotateleft(rbnode *rbn)
{
  rbnode *parent, *right;
  
  parent = rbn->parent;
  right = rbn->right;
  
  rbn->parent = right;
  rbn->right = right->left;
  rbn->right->parent = rbn;
  
  right->left = rbn;
  right->parent = parent;
  
  if (parent)
  {
    if (rbn == parent->left)  parent->left = right;
    else parent->right = right;
  }
}

void* getitemrbtree(rbtree *t, void *item)
{
  int result;
  rbnode *rbn = t->root;
  
  if (!rbn)  return NULL;
  
  while (rbn->item)
  {
    result = t->compar(item, rbn->item);
    if (result > 0)  rbn = rbn->right;
    else if (result < 0)  rbn = rbn->left;
    else  return rbn->item;
  }
  
  return NULL;
}

void* removeitemrbtree(rbtree *t, void *item)
{
  int result;
  void* rbitem = NULL;
  rbnode *rbn = t->root;
  
  while (!rbitem && rbn->item)
  {
    result = t->compar(item, rbn->item);
    if (result > 0)  rbn = rbn->right;
    else if (result < 0)  rbn = rbn->left;
    else  rbitem = rbn->item;
  }
  
  if (!rbitem)  return NULL;
  
  if (rbn == t->root)
  {
    if (!isleaf(rbn->left))  t->root = rbn->left;
    else if (!isleaf(rbn->right))  t->root = rbn->right;
    
    else t->root = NULL;
  }
  
  removerbnode(rbn);
  
  if (t->root)
  {
    while (t->root->parent)  t->root = t->root->parent;
    t->root->colour = BLACK;
  }
  
  return rbitem;
}

void removerbnode(rbnode *rbn)
{
  rbnode *rmnode;
  void *item;
  
  if (isleaf(rbn->left) || isleaf(rbn->right))
    rmnode = rbn;
  else
  {
    rmnode = findmaxrbnode(rbn->left);
    rbn->item = rmnode->item;
    //if (!rmnode->left)  
    rmnode->item = NULL;
  }
  
  deleterbnode(rmnode);
}

/**
 *  Must have at most one non-leaf child!
 */
void deleterbnode(rbnode *rbn)
{
  rbnode *child, *nodeptr, *sib;
  if (isleaf(rbn->left))  child = rbn->right;
  else if (isleaf(rbn->right))  child = rbn->left;
  else
  {
    fprintf(stderr, "Error! Attempting to call deleterbnode on a node with two non-leaf children!\n");
    exit(1);
  } 
  
  replacerbnode(rbn, child);
  
  if (rbn->colour == BLACK)
  {
    if (child->colour == RED)  child->colour = BLACK;
    else 
    {
      nodeptr = child;
      while (nodeptr->parent) // Fix this? nodeptr->parent?
      {
        //if (!nodeptr->parent)  break;
       
        sib = siblingrbnode(nodeptr);
          
        if (sib->colour == RED)
        {
          nodeptr->parent->colour = RED;
          sib->colour = BLACK;
          
          if (nodeptr == nodeptr->parent->left)  rotateleft(nodeptr->parent);
          else  rotateright(nodeptr->parent);
        }
        
        sib = siblingrbnode(nodeptr);
        
        if (nodeptr->parent->colour == BLACK && 
           sib->colour == BLACK && 
           sib->left->colour == BLACK &&
           sib->right->colour == BLACK)
        {
          sib->colour = RED;
          nodeptr = nodeptr->parent;
          
          // continue??
        }
        
        else if (nodeptr->parent->colour == RED &&
                 sib->colour == BLACK &&
                 sib->left->colour == BLACK &&
                 sib->right->colour == BLACK)
        {
          sib->colour = RED;
          nodeptr->parent->colour = BLACK;
          
          break;
        }
        
        else // sibling must be black... see wikipedia
        {
          if (nodeptr == nodeptr->parent->left &&
              sib->right->colour == BLACK &&
              sib->left->colour == RED)
          {
            sib->colour = RED;
            sib->left->colour = BLACK;
            rotateright(sib);
          }
          
          else if (nodeptr == nodeptr->parent->right &&
                   sib->left->colour == BLACK &&
                   sib->right->colour == RED)
          {
            sib->colour = RED;
            sib->right->colour = BLACK;
            rotateleft(sib);
          }
          
          sib = siblingrbnode(nodeptr);
          sib->colour = nodeptr->parent->colour;
          nodeptr->parent->colour = BLACK;
          
          
          if (nodeptr == nodeptr->parent->left)
          {
            sib->right->colour = BLACK;
            rotateleft(nodeptr->parent);
          }
          
          else
          {
            sib->left->colour = BLACK;
            rotateright(nodeptr->parent);
          }
          
          break;
        }
      }
    }
  }
  
  freerbnodeREC(rbn);
}

void replacerbnode(rbnode *rbn, rbnode *replacement)
{
  if (rbn->left == replacement)  rbn->left = NULL;
  else if (rbn->right == replacement)  rbn->right = NULL;
  else
  {
    fprintf(stderr, "Error! Attepting to replace a node with another that is not its child!\n");
    exit(1);
  }

  if (rbn->parent)
  {
    if (rbn->parent->left == rbn)  rbn->parent->left = replacement;
    else  rbn->parent->right = replacement;
  }
    
  replacement->parent = rbn->parent;
  rbn->parent = NULL;
  
}

rbnode* findmaxrbnode(rbnode *start)
{
  rbnode *rbptr = start;
  if (!start->item)  return NULL;
  while (rbptr->right->item)
    rbptr = rbptr->right;
  
  return rbptr;
}

int isleaf(rbnode *rbn)
{
  return (rbn->left == NULL && rbn->right == NULL);
  
}

void freerbtree(rbtree *t)
{
  
  if (t->root)  freerbnodeREC(t->root);
}

void freerbnodeREC(rbnode *rbn)
{
  if (rbn->left)  freerbnodeREC(rbn->left);
  if (rbn->right)  freerbnodeREC(rbn->right);
  
  free(rbn);
}

