/**
 * tree.c
 * A PHYLIP-style tree structure, where internal nodes are represented by a set
 * of node structures, one node per branch. See "Inferring Phylogenies" by
 * Joe Felsenstein for a detailed description
 * (c) 2009 JW Leigh
 */
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <stdbool.h>
#include "ds.h"

#define MAXNAMLEN 100
#define BRLENCHARS 100

treenode* inittreenode(treenode *tn, int index, int nleaves)
{
  tn->out = tn;
  tn->next = tn;
  tn->index = index;
  tn->brlen = 1.0;
  tn->label = NULL;
  tn->binrep = (treebipart*)malloc(sizeof(treebipart));
  
  if (index < nleaves)
    inittreebipart(tn->binrep, nleaves, index);
  else
    inittreebipart(tn->binrep, nleaves, -1);
  
  return tn;
}

void resettreenode(treenode *tn, int index, int nleaves)
{
  tn->out = tn;
  tn->next = tn;
  tn->index = index;
  tn->brlen = 1.0;
  tn->label = NULL;
  
  if (index < nleaves)
    resettreebipart(tn->binrep, nleaves, index);
  else
    resettreebipart(tn->binrep, nleaves, -1);
}

void freetreenode(treenode *tn)
{
  if (tn->binrep)  
  {
    freetreebipart(tn->binrep);
    free(tn->binrep);
  }
}

treebipart* inittreebipart(treebipart *tbp, int nleaves, int index)
{
  int i;
  
  tbp->array = (bool*)malloc(nleaves * sizeof(bool));
  tbp->nleaves = nleaves;
  for (i = 0; i < nleaves; i++)  tbp->array[i] = FALSE;
  if (index > 0)
    tbp->array[index] = TRUE;
  
  return tbp;
}

void resettreebipart(treebipart *tbp, int nleaves, int index)
{
  int i;
  
  if (nleaves != tbp->nleaves)
  {
    tbp->array = (bool*)realloc(tbp->array, nleaves * sizeof(bool));
    tbp->nleaves = nleaves;
  }
  
  for (i = 0; i < tbp->nleaves; i++)  tbp->array[i] = FALSE;
  
  if (index > 0)
    tbp->array[index] = TRUE;
  
}


void freetreebipart(treebipart *tbp)
{
  free(tbp->array);
}

treebipart* combinetreebipart(treebipart *dest, treebipart *other)
{
  int i;
  
  for (i = 0; i < dest->nleaves; i++)//{
    dest->array[i] = (dest->array[i] || other->array[i]);
  
  return dest;
}

treebipart* copytreebipart(treebipart *dest, treebipart *src)
{
  int i;
  
  inittreebipart(dest, src->nleaves, -1);
  for (i = 0; i < src->nleaves; i++)  dest->array[i] = src->array[i];
  
  return dest;
}

/**
 * Note that if leafnames are not sorted, the order of the nodearray element will NOT be the same 
 * 
 */
tree* inittree(tree *t, char **leafnames, int nleaves)
{
  int i;
  treenode *intern, *leaf;
  
  t->nleaves = nleaves;
  t->ninternal = 0;
  t->leafnames = leafnames;
  
  t->nodearray = (treenode**)malloc((2 * nleaves - 2) * sizeof(treenode*));
  
  for (i = 0; i < nleaves; i++)
  {
    leaf = (treenode*)malloc(sizeof(treenode));
    inittreenode(leaf, i, nleaves);
    leaf->label = leafnames[i];
    t->nodearray[i] = leaf;
  }
  
  qsort(t->nodearray, t->nleaves, sizeof(treenode*), cmptreenode);
  
  for (i = 0; i < nleaves; i++)  t->nodearray[i]->index = i;
  
  t->treestr = (char*)malloc(BUFSIZ * sizeof(char));
  t->root = t->nodearray[0];
  
  t->spareleaves = NULL;
  
  return t;
  
}

tree* freetree(tree *t)
{
  int i;
  treenode *tn, *tm;
  
  for (i = 0; i < (t->nleaves + t->ninternal); i++)  
  {
    tn = t->nodearray[i]->next;
    
    do
    {
      tm = tn->next;
      freetreenode(tn);
      free(tn);
      tn = tm;
    } while (tn != t->nodearray[i]);
  }
  
  free(t->nodearray);
  free(t->treestr);
}

tree* restring2tree(tree *t, char *trstr)
{
  bool inbrlen = FALSE, inname = FALSE;
  int i;
  int namlen = 0, blchars = 0;
  char *namptr = (char*)malloc((MAXNAMLEN + 1) * sizeof(char));
  char *blstr = (char*)malloc((BRLENCHARS + 1) * sizeof(char));
  double bl;
  char *charit;
  treenode *current = NULL;
  treenode *sparenodes = NULL, *sparetail;
  treenode *last, *next;
  
  for (i = 0; i < t->ninternal; i++)
  {
    current = t->nodearray[i + t->nleaves];
    do  
    {
      if (sparenodes)
      {
        sparetail->next = current;
        current->out = sparetail;
      }
      
      else
      {
        sparenodes = current;
        sparenodes->out = NULL;
      }
      
      sparetail = current;
      current = current->next;
      sparetail->next = NULL;
    } while (current != t->nodearray[i + t->nleaves]);
    
    t->nodearray[i + t->nleaves] = NULL;
  }
  //disconnect leaves as well
  t->ninternal = 0;
  for (i = 0; i < t->nleaves; i++)  t->nodearray[i]->out = t->nodearray[i]->next = t->nodearray[i];
  
  while (t->spareleaves)
  {
    t->nodearray[t->nleaves] = t->spareleaves;
    t->spareleaves = t->spareleaves->next;
    t->nodearray[t->nleaves]->out = t->nodearray[t->nleaves]->next = 
        t->nodearray[t->nleaves];
    t->nleaves++;
  }
  
  qsort(t->nodearray, t->nleaves, sizeof(treenode*), cmptreenode);
  
  for (i = 0; i < t->nleaves; i++)  t->nodearray[i]->index = i;

  charit = trstr;
  current = NULL;
  
  while (charit[0])
  {
    switch(charit[0])
    {
      /* ignore whitespace */
      case ' ':
      case '\n':
      case '\r':
      case '\t':
        //continue;
        break;
        
      case '(':
        if (current)
        {
          if (sparenodes)
          {
            current->next = sparenodes;
            sparenodes = sparenodes->next;
            if (sparenodes)  sparenodes->out = NULL;
	    resettreenode(current->next, current->index, t->nleaves);
          }
          
          else  
	  {
	    current->next = (treenode*)malloc(sizeof(treenode));
            inittreenode(current->next, current->index, t->nleaves);
	  }
          current = current->next;
          
          if (sparenodes)
          {
            current->out = sparenodes;
            sparenodes = sparenodes->next;
            if (sparenodes)  sparenodes->out = NULL;
	    resettreenode(current->out, t->ninternal + t->nleaves, t->nleaves);
          }
          
          else  
	  {
	    current->out = (treenode*)malloc(sizeof(treenode));
          
            inittreenode(current->out, t->ninternal + t->nleaves, t->nleaves);
	  }
	  
          current->out->out = current;
          current = current->out;
        }
        
        else
        {
          if (sparenodes)
          {
            current = sparenodes;
            sparenodes = sparenodes->next;
            if (sparenodes)  sparenodes->out = NULL;
	    resettreenode(current, t->ninternal + t->nleaves, t->nleaves);
          }
          else  
	  {
	    current = (treenode*)malloc(sizeof(treenode));
          
            inittreenode(current, t->ninternal + t->nleaves, t->nleaves);
	  }
        }
        
        t->nodearray[t->nleaves + t->ninternal] = current;
        t->ninternal++;
        break;
        
      case ')':
        if (inname)
        {
          inname = FALSE;
          namptr[namlen] = '\0';
          if (sparenodes)
          {
            current->next = sparenodes;
            sparenodes = sparenodes->next;
            if (sparenodes)  sparenodes->out = NULL;
	    resettreenode(current->next, current->index, t->nleaves);
          }
	  
          else  
	  {
	    current->next = (treenode*)malloc(sizeof(treenode));
          
	    inittreenode(current->next, current->index, t->nleaves);
	  }
	  
          current = current->next;
          if (!connectleaftree(t, current, namptr))
          {
            fprintf(stderr, "Error! Taxon %s in tree is not in list of taxa!\n", namptr);
            exit(1);
          }
        }
        current->next = t->nodearray[current->index];
        
        if (inbrlen)
        {
          inbrlen = FALSE;
          blstr[blchars] = '\0';
          bl = atof(blstr);
          if (!setbrlentreenode(current, bl))
          {
            fprintf(stderr, "Error! Attempting to set branch length on unconnected or NULL node!\n");
            exit(1);
          }
        }
        current = current->next->out;
        break;
        
      case ',':
        if (inname)
        {
          inname = FALSE;
          namptr[namlen] = '\0';
          if (sparenodes)
          {
            current->next = sparenodes;
            sparenodes = sparenodes->next;
            if (sparenodes)  sparenodes->out = NULL;
	    resettreenode(current->next, current->index, t->nleaves);
          }
	  
          else  
	  {
	    current->next = (treenode*)malloc(sizeof(treenode));
          
	    inittreenode(current->next, current->index, t->nleaves);
	  }
	  
          current = current->next;
          if (!connectleaftree(t, current, namptr))
          {
            fprintf(stderr, "Error! Taxon %s in tree is not in list of taxa!\n", namptr);
            exit(1);
          }
        }
        
        if (inbrlen)
        {
          inbrlen = FALSE;
          blstr[blchars] = '\0';
          bl = atof(blstr);
          if (!setbrlentreenode(current, bl))
          {
            fprintf(stderr, "Error! Attempting to set branch length on unconnected or NULL node!\n");
            exit(1);
          }
        }
        
        break;
        
      case ':':
        inbrlen = TRUE;
        blchars = 0;
        break;
        
      case ';':
        /* Silly hack! Stupid to need to free a node. */
        if (current->out == current)
        {
          last = current->next;
          while (last->next != current)  last = last->next;
          
          next = current->next;
          last->next = next;
          t->nodearray[next->index] = next;
	  freetreenode(current);
          free(current);
          current = next;
        }
        
        if (current->next->next == current && t->nleaves > 2)
        {
          if (!remtrivialnodetree(t, current))
          {
            fprintf(stderr, "Unknown error: Failed to remove trivial node.\n");
            exit(1);
          }
  
          freetreenode(current->next);
          free(current->next);
	  freetreenode(current);
          free(current);
          charit[1] = '\0';
        }
        
        break;
        
      default:
        if (inbrlen)
        {
          blstr[blchars] = charit[0];
          blchars++;
        }
        
        else if (inname)
        {
          namptr[namlen] = charit[0];
          namlen++;
        }
        
        else
        {
          inname = TRUE;
          namptr[0] = charit[0];
          namlen = 1;
        }
    }
    
    charit = &(charit[1]);
  }
  
  if (sparenodes)  free(sparenodes);
  
  free(namptr);
  free(blstr);
  
  t->root = t->nodearray[0];
  
  return t;
}

tree* string2tree(tree *t, char *trstr)
{
  bool inbrlen = FALSE, inname = FALSE;
  int i;
  int namlen = 0, blchars = 0;
  char *namptr = (char*)malloc((MAXNAMLEN + 1) * sizeof(char));
  char *blstr = (char*)malloc((BRLENCHARS + 1) * sizeof(char));
  double bl;
  char *charit;
  treenode *current = NULL;
  treenode *last, *next;
  
  charit = trstr;
  
  while (charit[0])
  {
    switch(charit[0])
    {
      /* ignore whitespace */
      case ' ':
      case '\n':
      case '\r':
      case '\t':
        //continue;
        break;
        
      case '(':
        if (current)
        {
          current->next = (treenode*)malloc(sizeof(treenode));
          inittreenode(current->next, current->index, t->nleaves);
          current = current->next;
          current->out = (treenode*)malloc(sizeof(treenode));
          inittreenode(current->out, t->ninternal + t->nleaves, t->nleaves);
          current->out->out = current;
          current = current->out;
        }
        
        else
        {
          current = (treenode*)malloc(sizeof(treenode));
          inittreenode(current, t->ninternal + t->nleaves, t->nleaves);
        }
        
        t->nodearray[t->nleaves + t->ninternal] = current;
        t->ninternal++;
        break;
        
      case ')':
        if (inname)
        {
          inname = FALSE;
          namptr[namlen] = '\0';
          current->next = (treenode*)malloc(sizeof(treenode));
          inittreenode(current->next, current->index, t->nleaves);
          current = current->next;
          if (!connectleaftree(t, current, namptr))
          {
            fprintf(stderr, "Error! Taxon %s in tree is not in list of taxa!\n", namptr);
            exit(1);
          }
        }
        current->next = t->nodearray[current->index];
        
        if (inbrlen)
        {
          inbrlen = FALSE;
          blstr[blchars] = '\0';
          bl = atof(blstr);
          
          /*if (bl < minbl)
          {
            next = current->next;
            
            collapsebranchtree(current);
            free(current->out);
            free(current);
            
            current = next->out;
            break;
          }*/
          
          if (!setbrlentreenode(current, bl))
          {
            fprintf(stderr, "Error! Attempting to set branch length on unconnected or NULL node!\n");
            exit(1);
          }
        }
        current = current->next->out;
        break;
        
      case ',':
        if (inname)
        {
          inname = FALSE;
          namptr[namlen] = '\0';
          current->next = (treenode*)malloc(sizeof(treenode));
          inittreenode(current->next, current->index, t->nleaves);
          current = current->next;
          if (!connectleaftree(t, current, namptr))
          {
            fprintf(stderr, "Error! Taxon %s in tree is not in list of taxa!\n", namptr);
            fprintf(stderr, "Tree string:\n%s", trstr);
            exit(1);
          }
        }
        
        if (inbrlen)
        {
          inbrlen = FALSE;
          blstr[blchars] = '\0';
          bl = atof(blstr);
          /*if (bl < minbl)
          {
            next = current->next;
            last = next;
            while (last->next != current)  last = last->next;
            
            collapsebranchtree(current);
            
            if (current->out != current)  free(current->out);
            free(current);
            
            current = last;
            while (current->next != next)  current = current->next;
            break;
          }*/
         
          if (!setbrlentreenode(current, bl))
          {
            fprintf(stderr, "Error! Attempting to set branch length on unconnected or NULL node!\n");
            exit(1);
          }
        }
        
        break;
        
      case ':':
        inbrlen = TRUE;
        blchars = 0;
        break;
        
      case ';':
        /* Silly hack! Stupid to need to free a node. */
        if (current->out == current)
        {
          last = current->next;
          while (last->next != current)  last = last->next;
          
          next = current->next;
          last->next = next;
          t->nodearray[next->index] = next;
	  
	  freetreenode(current);
          free(current);
          current = next;
        }
        
        if (current->next->next == current && t->nleaves > 2)
        {
          if (!remtrivialnodetree(t, current))
          {
            fprintf(stderr, "Unknown error: Failed to remove trivial node.\n");
            exit(1);
          }
  
          freetreenode(current->next);
	  free(current->next);
	  freetreenode(current);
          free(current);
        }
        
        charit[1] = '\0';
        break;
        
      default:
        if (inbrlen)
        {
          blstr[blchars] = charit[0];
          blchars++;
        }
        
        else if (inname)
        {
          namptr[namlen] = charit[0];
          namlen++;
        }
        
        else
        {
          inname = TRUE;
          namptr[0] = charit[0];
          namlen = 1;
        }
    }
    
    charit = &(charit[1]);
  }
  
  free(namptr);
  free(blstr);
  
  return t;
}

/**
 * Remove a node with only two "out" connections. Fix indices of nodes in 
 * t->nodearray. Return removed node.
 * If this is not a trivial node, do nothing and return NULL.
 */
treenode* remtrivialnodetree(tree *t, treenode *tn)
{
  treenode *out, *nextout, *nptr;
  int i;
  
  if (tn->next->next != tn || t->nleaves <= 2)  return NULL;
  
  nextout = tn->next->out;
  out = tn->out;
          
  if (!nextout || !out)
  {
    fprintf(stderr, "Unknown error: it seems there are more than two taxa, but root does not point \"out\" to enough.\n");
    exit(1);
  }
  
  if (tn->index != t->nleaves)
  {
    fprintf(stderr, "Trivial node has the wrong index. \n");
    exit(1);
  }
          
  out->out = nextout;
  nextout->out = out;
  out->brlen += nextout->brlen;
  nextout->brlen = out->brlen;
  
  
  t->ninternal--;
   
  for (i = t->nleaves; i < (t->nleaves + t->ninternal); i++)
  {
    t->nodearray[i] = t->nodearray[i+1];
    nptr = t->nodearray[i];
    nptr->index = i;
            
    while ((nptr = nptr->next) != t->nodearray[i])
      nptr->index = i;
  }
          
  t->nodearray[t->nleaves + t->ninternal] = NULL;
  
  return tn;
}

void collapseshortbranchestree(tree *t, double minbrlen)
{
  
  bool terminalbranch = false;
  
  if (t->root->next != t->root)
    collapseshortbranchestreeREC(t, t->root, minbrlen);
  
  else  terminalbranch = true;
  
  if (t->root->out->next != t->root->out)
    collapseshortbranchestreeREC(t, t->root->out, minbrlen);
  
  else  terminalbranch = true;
  
  if (! terminalbranch)
  {
    treenode *p = t->root->next;
    collapsebranchtree(t->root);
    t->nodearray[t->root->out->index] = NULL;
    freetreenode(t->root->out);
    free(t->root->out);
    freetreenode(t->root);
    free(t->root);
    t->root = p;
  }
    
  // fix nodearray:
  int nextempty;// = t->nleaves;
  int nextnode;// = t->nleaves + 1;
  int n = t->nleaves + t->ninternal;
  treenode *p;
  for (nextempty = t->nleaves; nextempty < (n - 1); nextempty++)   
  {
    if (t->nodearray[nextempty] == NULL)
    {
      // find the next non-empty cell
      nextnode = nextempty + 1;
      while (nextnode < n && t->nodearray[nextnode] == NULL) nextnode++;
      
      if (nextnode < n)
      {
        p = t->nodearray[nextnode];
        do 
        {
          p->index = nextempty;
          p = p->next;
        }  while (p != t->nodearray[nextnode]);
        
        t->nodearray[nextempty] = t->nodearray[nextnode];
        //t->ninternal = nextempty - t->nleaves;
        t->nodearray[nextnode] = NULL;
      }
    }
  }
  
  // TODO integrate this with function above
  int i;
  t->ninternal = 0;
  for (i = t->nleaves; i < n; i++)
    if (t->nodearray[i])  t->ninternal++;
}

void collapseshortbranchestreeREC(tree *t, treenode *tn, double minbrlen)
{
  treenode *p = tn->next;
  treenode *q;
  while (p != tn)
  {
    q = p->next;
    // don't collapse leaf branches
    if (p->out->next != p->out)
    {
      collapseshortbranchestreeREC(t, p->out, minbrlen);
      
      if (p->brlen < minbrlen)
      {
        collapsebranchtree(p);
        t->nodearray[p->out->index] = NULL;
        t->nodearray[p->index] = q;
        freetreenode(p->out);
        free(p->out);
        freetreenode(p);
        free(p);
        
      }
      
    }
     
    p = q;
    
  }
}


void collapsebranchtree(treenode *tn)
{
  if (tn->next == tn || tn->out->next == tn->out)
  {
    fprintf(stderr, "Cannot collapse a terminal branch.\n");
    exit(1);
  }
  
  // find the nodes before and after tn
  treenode *p;
  treenode *q = tn->next;
  
  while (q->next != tn)  q = q->next;
  
    
  p = tn->out->next;
  p->index = q->index;
  q->next = p;
  while (p->next != tn->out)  
  {
    p = p->next;
    p->index = q->index;
  }

  p->next = tn->next;
  
  // clean up 
  tn->out->next = tn->out;
  tn->next = tn;
}


bool connectleaftree(tree *t, treenode *tn, char *leafname)
{
  treenode *leaf = findnodetree(t, leafname, 0, t->nleaves - 1);
  if (!leaf)  return FALSE;
  
  leaf->out = tn;
  tn->out = leaf;
  
  return TRUE;
}

bool setbrlentreenode(treenode *tn, double brlen)
{
  if (!tn || !tn->out)  return FALSE;
  
  tn->brlen = brlen;
  tn->out->brlen = brlen;
  
  return TRUE;
}

char* makestringtree(tree *t)
{
  treenode *tn;
  t->treestr[0] = '(';
  t->treestr[1] = '\0';
  
  tn = t->root->out->next;
  makestringtreeREC(t, t->root);
  
  while (tn != t->root->out)
  {
    strcat(t->treestr, ",");
    makestringtreeREC(t, tn->out);
    tn = tn->next;
  }
  
  strcat(t->treestr, ");");
  
  return t->treestr;
}

void makestringtreeREC(tree *t, treenode *tn)
{
  treenode *ptr;
  int trstrlen;
  
  if (tn->next == tn)
  {
    if (tn->label)
    {
      trstrlen = strlen(t->treestr);
      sprintf(&(t->treestr[trstrlen]), "%s:%.2f", tn->label, tn->brlen); 
    }
    
    else  strcat(t->treestr, "ERR:0.0");
  }
  
  else
  {
    strcat(t->treestr, "(");
    ptr = tn->next;
    
    while (ptr != tn)
    {
      makestringtreeREC(t, ptr->out);
      ptr = ptr->next;
      if (ptr != tn)  strcat(t->treestr, ",");
    }
    
    trstrlen = strlen(t->treestr);
    sprintf(&(t->treestr[trstrlen]), "):%.2f", tn->brlen);
  }
}

void printtree(tree *t)
{
  printf("%s\n", makestringtree(t));
}

hashtable* countbipartstree(tree *t, hashtable *bipartcnts)//double *bipartcnts)
{
  treenode *tn;
  
  t->root = t->nodearray[0];
  tn = t->root->out->next;
  
  while (tn != t->root->out)
  {
    countbipartstreeREC(t, tn->out, bipartcnts);
    tn = tn->next;
  }
  
  return bipartcnts;
}

void countbipartstreeREC(tree *t, treenode *tn, hashtable *bipartcnts)
{
  treenode *ptr;
  treebipart *key;
  double *counter;
  int i;
  
  if (tn->binrep->nleaves != t->nleaves)
  {
    fprintf(stderr, "Wrong number of leaves for bipart: %d\n", tn->binrep->nleaves);
    exit(1);
  }
  /*if (tn->next == tn)
  {
    if (!tn->binrep)
    {
      tn->binrep = (treebipart*)malloc(sizeof(treebipart));
      inittreebipart(tn->binrep, t->nleaves, tn->index);
    }
  }*/
  
  if (tn->next != tn)//else
  {
    ptr = tn->next;
    /*if (!tn->binrep)
    {
      tn->binrep = (treebipart*)malloc(sizeof(treebipart));
      inittreebipart(tn->binrep, t->nleaves, -1);
    }*/
    
    //else  
    if (tn->index < tn->binrep->nleaves)
      resettreebipart(tn->binrep, tn->binrep->nleaves, tn->index);
    else
      resettreebipart(tn->binrep, tn->binrep->nleaves, -1);

    while (ptr != tn)
    {
      countbipartstreeREC(t, ptr->out, bipartcnts);
      combinetreebipart(tn->binrep, ptr->out->binrep);
      
      ptr = ptr->next;
    }
    
    counter = (double*)gethashtable(bipartcnts, tn->binrep);
    if (counter)  (*counter)++;
    
    else
    {
      counter = (double*)malloc(sizeof(double));
      *counter = 1.0;
      key = (treebipart*)malloc(sizeof(treebipart));
      inserthashtable(bipartcnts, copytreebipart(key, tn->binrep), counter);
    }
  }
}

void freevoidtreebp(void *vtbp)
{
  treebipart *tbp = (treebipart*)vtbp;
  
  freetreebipart(tbp);
}

void printbipartstree(tree *t)
{
  hashtable *bipartstab = (hashtable*)malloc(sizeof(hashtable));
  int i, charcount;
  treebipart *bpptr;
  
  inithashtable(bipartstab, hashtreebipart, cmptreebipart, freevoidtreebp, NULL, HASHSIZE);
  
  printf("Taxa:");
  for (i = 0; i < t->nleaves; i++)
  {
    if (!(i % 10))  printf("\n");
    printf("%d. %s\n", i+1, t->nodearray[i]->label);
  }
  printf("\nBipartitions:\n");
  
  countbipartstree(t, bipartstab);
  resetiterhashtable(bipartstab);
  
  while ((bpptr = (treebipart*)nextkeyhashtable(bipartstab)) != NULL)
  {
    for (i = 0; i < bpptr->nleaves; i++)
    {
      if (i && !(i % 10))  printf(" ");
      if (bpptr->array[i])  printf("*");
      else printf("-");
    }
         
    printf("\n");
    
    free(gethashtable(bipartstab, bpptr));
    freetreebipart(bpptr);
    free(bpptr);
  }
  
  freehashtable(bipartstab);
  free(bipartstab);
}

static int cmptreenode(const void *node1, const void *node2)
{
  treenode *tn = *(treenode**)node1;
  treenode *tm = *(treenode**)node2;
  
  return strcmp(tn->label, tm->label); 
}

int hashtreebipart(const void *p1)
{
  treebipart *binrep = (treebipart*)p1;
  int i;
  int hashval = 0;
  
  for (i = 0; i < binrep->nleaves && (1 << i) < (28 * HASHSIZE); i++)
    if (binrep->array[i])  hashval += (1 << i);
  
  hashval = (hashval / 28) % HASHSIZE;
  
  return hashval;
}

/**
 * If the two treebiparts have different numbers of leaves, returns neg if p1 
 * has fewer leaves, pos if p2 has fewer.
 * Otherwise, returns 0 if the two are the same, neg if p1 is represented by a
 * a larger binary number, pos if p2 has a smaller bin rep
 */
int cmptreebipart(const void *p1, const void *p2)
{
  int i;
  treebipart *tbp1 = *(treebipart* const*)p1;
  treebipart *tbp2 = *(treebipart* const*)p2;
  
  if (tbp1->nleaves < tbp2->nleaves)  return -1;
  if (tbp2->nleaves < tbp1->nleaves)  return 1;
   
  for (i = tbp1->nleaves - 1; i >= 0; i--)
  {
    if (tbp1->array[i] != tbp2->array[i])
    {
      if (tbp1->array[i] == 1)  return -1;
      return 1;
    }
  }
  
  return 0;
}

/**
 * Search for treenode with the given label in the tree's nodearray. Uses binary search algo.
 */
treenode* findnodetree(tree *t, char *label, int startidx, int endidx)
{
  int mid;
  treenode *tn;
  
  if (startidx == endidx)
  {
    if (strcmp(label, t->nodearray[startidx]->label) == 0)  return t->nodearray[startidx];
    
    else  return NULL;
  }
  
  mid = (endidx - startidx) / 2 + startidx;
  tn = findnodetree(t, label, startidx, mid);
  if (tn)  return tn;
  else return findnodetree(t, label, mid + 1, endidx);
}

void pruneleaftree(tree *t, char *label)
{
  treenode *leaf = findnodetree(t, label, 0, t->nleaves - 1);
  treenode *out, *nodeptr, *nextout, *prevout;
  int outidx = -1, leafidx;
  int i;
  int numnodes;
  
  if (leaf)
  {
    leafidx = leaf->index;
    out = leaf->out;
    
    if (t->spareleaves)  leaf->next = t->spareleaves;
      
    else  leaf->next = NULL;
      
    t->spareleaves = leaf;

    //free(leaf);
    
    nodeptr = out->next;
    
    /* leaf's parent was bifurcating, delete */
    if (nodeptr->next->next == out)
    {
      outidx = out->index;
      nextout = nodeptr->out;
      prevout = nodeptr->next->out;
      nextout->brlen += prevout->brlen;
      prevout->brlen = nextout->brlen;
      nextout->out = prevout;
      prevout->out = nextout;
      freetreenode(out);
      free(out);
      freetreenode(nodeptr->next);
      free(nodeptr->next);
      freetreenode(nodeptr);
      free(nodeptr);
    }
    
    else
    {
      while (nodeptr->next != out)  nodeptr = nodeptr->next;
      
      nodeptr->next = nodeptr->next->next;
      freetreenode(out);
      free(out);
    }
    
    numnodes = t->nleaves + t->ninternal - 1;
    if (outidx > 0) numnodes--;
    
    for (i = leafidx; i < numnodes; i++)
    {
      if (i < (outidx - 1) || outidx < 0)
        t->nodearray[i] = t->nodearray[i+1];
      else
        t->nodearray[i] = t->nodearray[i+2];
      
      nodeptr = t->nodearray[i];
      do
      {
        nodeptr->index = i;
        nodeptr = nodeptr->next;
      } while  (nodeptr != t->nodearray[i]);
    }
    t->nleaves--; 
    t->nodearray[t->nleaves + t->ninternal] = NULL;
   
    
    if (outidx > 0)
    {
      t->ninternal--;
      t->nodearray[t->nleaves + t->ninternal] = NULL;
    }
    
    t->root = t->nodearray[0];
  }
  
  else
  {
    fprintf(stderr, "Error! No node with label %s found!\n", label);
    exit(1);
  }
}

void fixleavestree(tree *t)
{
  // iterate over leaf nodes
  // if a leaf->out points to itself, it's not valid... increment the number of "shift" back indices
  // once we get to nleaves + ninternal - nshift, make
  // remaining cells NULL
  // nleaves -= nshift
  // move t->root to (maybe new) t->nodearray[0]
  
  int i;
  int nshift = 0;
  treenode *nodeptr;
  
  for (i = 0; i < t->nleaves; i++)
  {
    if (t->nodearray[i]->out == t->nodearray[i])
    {
      if (t->spareleaves)  t->nodearray[i]->next = t->spareleaves;
      
      else  t->nodearray[i]->next = NULL;
      
      t->spareleaves = t->nodearray[i];
      
      nshift++;
    }
    
    else
    {
      t->nodearray[i]->index = i - nshift;
      //resettreebipart(t->nodearray[i]->binrep, t->nleaves, i - nshift);
      t->nodearray[i - nshift] = t->nodearray[i];
    }
  }
    
  for (i = t->nleaves; i < t->nleaves + t->ninternal; i++)
  {
    
    nodeptr = t->nodearray[i];
    //resettreebipart(nodeptr->binrep, t->nleaves, -1);
    do
    {
      nodeptr->index = i - nshift;
      nodeptr = nodeptr->next;
    } while  (nodeptr != t->nodearray[i]);
    
    t->nodearray[i - nshift] = t->nodearray[i];
    
    if (i >= t->nleaves + t->ninternal - nshift)
      t->nodearray[i] = NULL;
  }
  
  t->nleaves -= nshift;
  
  t->root = t->nodearray[0];
  
  for (i = 0; i < t->nleaves + t->ninternal; i++)
  {
    /*if (t->nleaves != 50)
    {
      fprintf(stderr, "tree leaves: %d ; bipart leaves: %d ; ", t->nleaves, t->nodearray[i]->binrep->nleaves);
    }*/
    
    if (i < t->nleaves)
      resettreebipart(t->nodearray[i]->binrep, t->nleaves, i);
    else
    {
     nodeptr = t->nodearray[i];

      do
      {
        resettreebipart(nodeptr->binrep, t->nleaves, -1);
        nodeptr = nodeptr->next;
      } while  (nodeptr != t->nodearray[i]);
    
    }
    
    /*if (t->nleaves != 50)
    {
      fprintf(stderr, "bipart leaves now: %d ; root leaves: %d\n", t->nodearray[i]->binrep->nleaves, t->root->binrep->nleaves);
    }*/
  }
}


